// ppmrectlist.cc
// Copyright (C) 2013 Michael Rose

// ============================================================================
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ============================================================================

// This program extract a list of rectangles from a PPM file
// suitable to be used in the program 'djvumake'.
// All pictures read and written by this program must be in raw PPM format.
// The program is testet with Linux but should also run on other platforms
// with only minor modifications. No special libraries are used.
// To compile use something like:
//
// $ g++ -o ppmrectlist ppmrectlist.cc
//
// A brief help text is printed, when wrong command syntax is used or simply
// by typing:
//
// $ ppmrectlist -h
//
// Just call the program like
//
// $ ppmrectlist picture.ppm > rectlist.txt
//
// Black rectangles are ignored, because black is the default color in DJVU.

// ============================================================================

// Include files; no other libraries needed than 'libgpp' and 'ppmroselib.o':

#include "ppmroselib.h"

// ============================================================================

// ============================================================================
// Global parameters/options.
// ============================================================================

struct ParameterRectList : Parameter
{
  // File names:

  const char *sourcePicName;         // Skewed input image.
  const char *destPicName;           // Unwarped output image.

  ParameterRectList() : Parameter("ppmrectlist") { }

  void Define();
  void Check();
};

// ----------------------------------------------------------------------------

// Program parameters are global for ease of use:

static ParameterRectList param;

int main(int argc, char *argv[]) { return Handler(argc, argv, param); }

// ----------------------------------------------------------------------------

void ParameterRectList::Define()
{
  AddString("-i", sourcePicName, 0, "(inpname)",
            "Set input file name");
  AddString("-o", destPicName, 0, "(stdout)",
            "Set output file name");
  ExtraUsage(
    "Simple usage:  ppmrectlist picture.ppm > rectlist.txt");
}

// ----------------------------------------------------------------------------

// Special argument checking:

void ParameterRectList::Check()
{
  if (!sourcePicName)  sourcePicName = inpName ? inpName : "";
  else if (inpName)    Usage(1);
  if (!destPicName)    destPicName = "";
}

// ============================================================================

// Type for image data:

struct Picture : Image
{
  void Write(const char *filename, FILE *fh=0) {
    Image::Write(filename, fh, param.prgName); }
};

// ============================================================================

// Type for rectangle lists:

struct Rect {
  int      x, y, w, h;
  pixel_t  r, g, b;

  void Set(int x, int y, int w, int h, pixel_t *raw) {
    this->x = x;  this->y = y;  this->w = w;  this->h = h;
    r = *raw++;  g = *raw++;  b = *raw++; }
  void Write(FILE *fh, long depth);
};

// ----------------------------------------------------------------------------

void Rect::Write(FILE *fh, long depth)
{
  int rr = r, gg = g, bb = b;
  if (depth != -1) {
    double scale = 255.0 / double(depth);
    rr = int(scale * double(r));
    gg = int(scale * double(g));
    bb = int(scale * double(b)); }
  fprintf(fh, "#%02x%02x%02x:%d,%d,%d,%d",
          rr, gg, bb, x, y, w, h);
}

// ============================================================================

// Rectangle list:

struct RectList {
  Picture& pic;

  long  width, height;
  int  *start;

  long  nlst, mlst;
  Rect *lst;

  RectList(Picture& pic) : pic(pic) { start = 0;  Reset(); }
  ~RectList()                       {             Reset(); }

  void Reset(int init=0);
  void NewRect(int x, int y, int w, int h, pixel_t *raw);
  void Create();
  void Write(const char *name, FILE *fh=0);
};

// ----------------------------------------------------------------------------

void RectList::Reset(int init)
{
  if (start) { delete start;  delete lst; }
  start = 0;
  lst   = 0;
  width = height = nlst = mlst = 0;
  if (init) {
    int k;
    width  = pic.width;
    height = pic.height;
    start  = new int[width];
    for (k=0; k < width; ++k)  start[k] = 0;
    lst = new Rect[mlst = 2]; }
}

// ----------------------------------------------------------------------------

void RectList::NewRect(int x, int y, int w, int h, pixel_t *raw)
{
  Rect *cur;
  if (raw[0] || raw[1] || raw[2]) {
    if (nlst >= mlst) {
      cur = new Rect[mlst <<= 1];
      for (int k=0; k < nlst; ++k)  cur[k] = lst[k];
      delete[] lst;
      lst = cur; }
    cur = lst + nlst++;
    // DJVUMAKE expects rectlist with origin in top bottom corner:
    cur->Set(x, height - y - h, w, h, raw); }
}

// ----------------------------------------------------------------------------

void RectList::Create()
{
  long     x, y, w, h, k, span;
  pixel_t *rawX, *rawY, *recX, *recY;
  Reset(1);
  span = 3 * width;
  for (y=0,rawY=pic.pixel; y < height; ++y,rawY+=span) {
    for (x=0,rawX=rawY; x < width; x+=w,rawX+=3*w) {
      if (start[x] <= y) {
        for (w=1,recX=rawX+3; x+w < width; ++w) {
          if (start[x+w] > y || rawX[0] != *recX++ ||
              rawX[1] != *recX++ || rawX[2] != *recX++)  break; }
        for (h=1,recY=rawX+span; y+h < height; ++h,recY+=span) {
          for (k=0,recX=recY; k < w; ++k) {
            if (rawX[0] != *recX++ || rawX[1] != *recX++ || rawX[2] != *recX++)
              break; }
          if (k < w)  break; }
        for (k=0; k < w; ++k)  start[x+k] = y+h;
        NewRect(x, y, w, h, rawX); }
      else  w = 1; } }
}

// ----------------------------------------------------------------------------

void RectList::Write(const char *filename, FILE *fh)
{
  long depth = pic.depth;
  if (fh != stdout)  fh = fopen(filename, "w");
  if (!fh)  Error("Couldn't write rectangle list file '%s'", filename);
  for (long k=0; k < nlst; ++k)  lst[k].Write(fh, depth);
  if (fh != stdout)  fclose(fh);
}

// ============================================================================

// Main routine:

void Main() {
  Picture  pic;
  RectList reclist(pic);
  if (*param.sourcePicName)  pic.Read(param.sourcePicName);
  else                       pic.Read("stdin", stdin);
  // Create the rectangle list:
  reclist.Create();
  if (*param.destPicName)  reclist.Write(param.destPicName);
  else                     reclist.Write("stdout", stdout);
}
